/*****************************************************************************
 * Secure Locate v1.3                                                        *
 * October 9, 1998                                                           *
 * Programmed by: Kevin Lindsay                                              *
 * Copyright (c) 1998 Kevin Lindsay & Netnation Communications Inc.          *
 *                                                                           *
 * NetNation Communications:  http://www.netnation.com/                      *
 * Secure Locate: ftp://ftp.mkintraweb.com/pub/linux/slocate/                *
 *                                                                           *
 * Report any Bugs to: klindsay@mkintraweb.com                               *
 *                                                                           *
 * Even though Secure Locate was written from scratch, some recognition goes *
 * to James A. Woods <jwoods@adobe.com> for writing the original GNU locate. *
 *****************************************************************************/

/* Secure Locate -- search database for filenames that match patterns without
 *                  showing files that the user using slocate does not have
 *                  access to.
 *    Copyright (C) 1994 Free Software Foundation, Inc.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2, or (at your option)
 *    any later version.
 * 
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 * 
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <netinet/in.h>
#include <grp.h>

#include "statmalloc.h"
#include "link.h"

/* GLOBALS */

#define VERSION "Secure Locate v1.3 - Released October 14, 1998\n"
#define SLOCATEDB "/var/lib/slocate/slocate.db"
#define TMPSLOCATEDB "/var/lib/slocate/slocate.tmp"
#define SLOCATEDB_DIR "/var/lib/slocate/"
#define GRPFILE "/etc/group"
#define SLOC_ESC 0x80
#define SLOC_GRP "slocate"
#define SLOC_UID 0
#define MIN_BLK 64

char *EXCLUDE_DIR;
int EXCLUDE=0;
char *progname;
char *tmp_path=NULL;
char *cur_dir=NULL;
short fr_num=0;
int t_num=0;
uid_t UID;
gid_t GID;
int first=1;
char slevel='1';
unsigned short SLOC_GID=65534; /* Seems to be the default nobody gid */

/* Usage */

void
usage() 
{
   printf("Secure Locate v1.3\n"
          "Copyright (c) 1998 Kevin Lindsay & Netnation Communications Inc.\n"
          "\n"
          "usage: %s [options] <search string>\n\n"
          "   Options:\n"
          "   -u                 - Create slocate database starting at path /.\n"
          "   -U <dir>           - Create slocate database starting at path <dir>.\n"          
          "   -e <dir1,dir2,...> - Exclude directories from slocate database.\n"
          "   -l <[01]>          - Security level. \n"
          "                           0 turns security checks off. This will make\n"
          "                             searchs faster.\n"
          "                           1 turns security checks on. This is the default.\n"
          "\n"
          "Author: Kevin Lindsay\n"
          "Bugs: klindsay@mkintraweb.com\n"
          "\n",progname);
   
   exit(0);
}

/* My fchdir() so that I can do cleaner error checking */
int
MYfchdir(int fd, char *filepath, char *func)
{   
   if (fchdir(fd) == -1) {
      fprintf(stderr,"%s: %s: fchdir: %s: %s\n",progname,func,filepath,strerror(errno));
      return(0);
   }
   return(1);
}

/* My chdir() so that I can do cleaner error checking */
int
MYchdir(char *filepath, char *func)
{
   if (chdir(filepath) == -1) {
      fprintf(stderr,"%s: %s: chdir: %s: %s\n",progname,func,filepath,strerror(errno));
      return(0);
   }
   return(1);
}

static void
put_short (int c, FILE *fp)
{
   putc (c >> 8, fp);
   putc (c, fp);
}

static int
get_short (FILE *fp)
{
   register short x;
   
   x = fgetc (fp);
   return (x << 8) | (fgetc (fp) & 0xff);
}

/* Make sure that the DB Directory exists */

int
check_dir()
{
	   int res=1;
	   struct stat tmpstat;
	
	   if (lstat(SLOCATEDB_DIR,&tmpstat) == -1) {
			      fprintf(stderr,"%s: error accessing DB Directory: %s : %s\n",progname,SLOCATEDB_DIR,strerror(errno));
			      res=0;
		}
	
	   return(res);
}

/* Get the GID for group slocate */

unsigned short
get_gid()
{
   unsigned short GGID=SLOC_GID;
	struct group *grpres;
	
	if ((grpres = getgrnam(SLOC_GRP)) == NULL) {		
      fprintf(stderr,"%s: Could not find the group: %s in the /etc/group file.\n",progname,SLOC_GRP);
      fprintf(stderr,"This is a result of the group missing or a corrupted group file.\n");
      exit(1);
	}

	GGID = grpres->gr_gid;
   
   return GGID;
}

/* Parse Exclude Command */

char *
parse_exclude(char *estr)
{
   char *excludestr=NULL;
   char *ptr1;
   char *ptr2=NULL;
   int grow=1;
   int elen;
   
   EXCLUDE=1;

   excludestr = (char *)malloc(1);
   excludestr[0] = '\0';
   
   elen = strlen(estr);

   ptr1 = estr;
   ptr2 = ptr1;

   while (ptr2[0] != '\0')  {

      if ((ptr2 = strchr(ptr1,',')) == NULL) {
         ptr2 = strchr(ptr1,'\0');
      }

      grow += (ptr2-ptr1)+2;
      excludestr = realloc(excludestr,grow);		
      strcat(excludestr,"*");
      strncat(excludestr,ptr1,ptr2-ptr1);
      strcat(excludestr,"*");
      if (ptr2[0] == ',')
          ptr1 = ptr2+1;
   }
  
	printf("E: %s\n",excludestr);
	exit(1);
   return (excludestr);
}

int
match_exclude(char *path,char *filedir) {
   int res=0;
   char *newstr;
   
   newstr = statmalloc(strlen(path)+strlen(filedir)+3);
   newstr[0] = '\0';
   strcat(newstr,"*");
   strcat(newstr,path);
   strcat(newstr,filedir);   
   strcat(newstr,"*");
   
   if (strstr(EXCLUDE_DIR,newstr) != NULL)       
       res=1;
   
   freemalloc(newstr);
   
   return res;
}

/* Load Dir Contents into Linked List */

dir_item *
load_dir(int dir_fd, dir_item *listptr)
{
   DIR *recdir;
   struct dirent *filename = NULL;
   struct stat statres;
   int top=0;

   if (MYfchdir(dir_fd,listptr->name,"LD_1") == -1) return(NULL);
   
   if ((recdir = opendir(".")) == NULL) {
      fprintf(stderr,"%s: Opendir: %s: %s\n",progname,listptr->name,strerror(errno));
      return(NULL);
   }

   /* Open Current Directory 
   if ((filename = readdir(recdir)) == NULL) {
      fprintf(stderr,"%s: Readdir Error: Load List: %s: %s\n",progname,listptr->name,strerror(errno));
      closedir(recdir);
      return(NULL);
   }

   seekdir(recdir,telldir(recdir)); */

   while ((filename = readdir(recdir)) != NULL) {

      while (!strcmp(filename->d_name,".") || !strcmp(filename->d_name,"..")) {
         top=1;
         seekdir(recdir,telldir(recdir));
         if ((filename = readdir(recdir)) == NULL)
             break;
      }
      
      if (top) {

         listptr = add_right(listptr);
         listptr = listptr->right;
         
         /* If directory is emtpy then add a space to listptr */         
         if (filename == NULL) {
            listptr->name = statmalloc(1);
            listptr->name[0] = '\0';
            listptr->empty = 1;
            listptr->type = 0;
            listptr->root = 0;
            listptr->name_len = 1;
            closedir(recdir);
            return(listptr);            
         }
         
         top=0;
      } else {
         listptr = add_down(listptr);
         listptr = listptr->down;
      }
      
      
      /* Add information from filenames->d_name to listptr */
      listptr->name = statmalloc(strlen(filename->d_name)+1);
      listptr->name[0] = '\0';
      strcat(listptr->name,filename->d_name);
      listptr->name_len = strlen(listptr->name);      

      if (lstat(filename->d_name,&statres) == -1) {
         fprintf(stderr,"%s: lstat: %s: %s\n",progname,filename->d_name,strerror(errno));
      }
      
      if (S_ISDIR(statres.st_mode))
          listptr->type=1;
      else
          listptr->type=0;
      
      listptr->root = 0;

      seekdir(recdir,telldir(recdir));      
   }
   
   closedir(recdir);
   
   return(listptr);
}

/* Get Current Directory */
char *
get_cur_dir(dir_item *listptr, char *dirstr)
{
   
   dir_item *dir_ptr;
   char **dirarray;
   int items=0;
   int i;
   int grow;
   

   if (cur_dir != NULL) {
      free(cur_dir);
   }

   dir_ptr = listptr;
   
   while (!dir_ptr->root) {
      dir_ptr=dir_ptr->left;
      items++;
   }

   dirarray = calloc(items,sizeof(char *));
   
   dir_ptr=listptr->left;
   
   for (i = items-1; i != -1; i--) {
      dirarray[i] = malloc(strlen(dir_ptr->name)+1);
      dirarray[i][0] = '\0';
      strcat(dirarray[i],dir_ptr->name);

      dir_ptr=dir_ptr->left;
   }

   cur_dir = malloc(1);
   cur_dir[0] = '\0';
   grow=1;
   
   for (i = 0; i != items; i++) {
      grow+=strlen(dirarray[i])+1;
      cur_dir = realloc(cur_dir,grow);
      strcat(cur_dir,dirarray[i]);
      if ((strcmp(dirstr,"/") != 0 && i == 0) || i > 0)
          strcat(cur_dir,"/");
   }

   for (i = 0; i != items; i++) {
      free(dirarray[i]);
   }
   
   free(dirarray);

   
   return(cur_dir);
}

/* FRCODE - Incremental Encoding algorithm */

int
frcode(char *dir_path, char *filename, struct stat statres)
{
   int res=0;
   char *cur_path;
   char *new_path;
   char *ptr1;
   char *ptr2;

   int i;
   FILE *fd;   
   
   cur_path = malloc(strlen(dir_path)+strlen(filename)+1);
   cur_path[0]='\0';
   strcat(cur_path,dir_path);
   strcat(cur_path,filename);
   
   if (tmp_path != NULL) {
      ptr1=cur_path;
      ptr2=tmp_path;
      
      for (i=0; ptr1[0] == ptr2[0] && (ptr1[0] != '\0' && ptr2[0] != '\0') ; i++) {
         ptr1++;
         ptr2++;
      }

      fr_num = i - t_num;
         
      t_num = t_num + fr_num;            
      
      ptr2 = strchr(ptr1,'\0');
      new_path=malloc((ptr2-ptr1)+1);
      new_path[0] = '\0';
      strncat(new_path,ptr1,ptr2-ptr1);         

   } else {
      new_path = malloc(strlen(cur_path)+1);
      new_path[0] = '\0';
      strcat(new_path,cur_path);
      t_num = strlen(new_path);
   }
   
   if (tmp_path != NULL)
       free(tmp_path);
   
   tmp_path = malloc(strlen(cur_path)+1);
   tmp_path[0] = '\0';
   strcat(tmp_path,cur_path);
   
   fd = fopen(TMPSLOCATEDB,"a");
   if (fr_num > 127 || fr_num < -127) {
      putc (SLOC_ESC, fd);
      put_short (fr_num, fd);
   } else
		 putc(fr_num,fd);
	
   fputs(new_path,fd);
   putc('\0',fd);

   fclose(fd);
   
   if (first) {
      chmod(TMPSLOCATEDB,00640);
      chown(TMPSLOCATEDB,UID,SLOC_GID);
      first=0;
   }
   
   free(new_path);
   free(cur_path);
   
   return res;
}

/* Create Database */

int
create_db(char *dirstr)
{
   int res=0;
   struct stat statres;
   int dir_fd;
   dir_item *listptr=NULL;
   dir_item *tmpptr=NULL;
   int load;
   int jump=0;
   int list_down=0;
   FILE *fd;

   if (UID != SLOC_UID) {
      fprintf(stderr,"%s: You are not authorized to create an slocate database!\n",progname);
      return(0);
   }
   
   fd = fopen(TMPSLOCATEDB,"w");
   putc(slevel,fd);
   fclose(fd);
   
   if (dirstr) {
      /* Make sure there are no problems accessing the source directory */
      if (lstat(dirstr,&statres) == -1) {
         fprintf(stderr,"%s: lstat: d%s/: %s\n",progname,dirstr,strerror(errno));
         return(-1);
      }
   }
   
   /* Initialize 2D Linked List */
   listptr = init_2D_list();

   if (dirstr) {
       if (dirstr[strlen(dirstr)-1] == '/') dirstr[strlen(dirstr)-1] = '\0';
   } else {
      dirstr = malloc(2);
      dirstr[0]='/';
      dirstr[1]='\0';
   }

   
   listptr->name = statmalloc(strlen(dirstr)+1);
   listptr->name[0] = '\0';
   strcat(listptr->name,dirstr);
   listptr->st_uid = statres.st_uid;
   listptr->st_gid = statres.st_gid;
   listptr->st_mode = statres.st_mode;
   listptr->root = 1;
   listptr->empty=0;
   
   
   /* Open Source Directory */
   if ((dir_fd = open(dirstr,O_RDONLY)) == -1) {
      fprintf(stderr,"%s: open: %s: %s\n",progname,dirstr,strerror(errno));
      return(-1);
   }
   
   /* Change CWD to source dir */
   if (!MYfchdir(dir_fd,dirstr,"CDB_1")) {
      close(dir_fd);
      return(0);
   }
   
   if (EXCLUDE && match_exclude(dirstr,""))
       load = 0;
   else
       load = 1;

   while (!jump) {
      if (load) {
         /* Load directory contents into linked list */
         listptr = load_dir(dir_fd,listptr);
         load=0;
         listptr=listptr->left->right;
      }
      
      if (!listptr->empty) {

         tmpptr = listptr;
         if (list_down) {
            listptr=listptr->down;
         }
         list_down=0;         
         while (listptr != NULL) {
            lstat(listptr->name,&statres);
            if (!EXCLUDE || (EXCLUDE && !match_exclude(get_cur_dir(listptr,dirstr),listptr->name))) {               
               if (statres.st_uid < 65535)
                   frcode(get_cur_dir(listptr,dirstr),listptr->name,statres);               
            }
            if (!listptr->type) {
               listptr=listptr->down;
            } else {
               load=1;
               break;
            }
         }
      } else {
         if (!MYchdir("..","create_db")) {
            close(dir_fd);
            return(1);
         }         
         listptr=listptr->left;
         free_right(listptr);
         tmpptr=listptr;
         listptr=listptr->down;
         if (listptr != NULL)
             load = listptr->type;
         list_down=0;
      }
      
      if (listptr == NULL) {
         
         while (listptr == NULL) {
            listptr=tmpptr;
            if (listptr->root) {
               jump=1;
               break;
            }
            listptr=listptr->left;
            
            if (!MYchdir("..","CDB_2")) {
               close(dir_fd);
               return(1);
            }
            
            close(dir_fd);
            
            if ((dir_fd = open(".",O_RDONLY)) == -1) {
               fprintf(stderr,"%s: open: CDB_3: %s\n",progname,strerror(errno));
               return(1);
            }
            
            tmpptr=listptr;
            free_right(listptr);
            listptr=listptr->down;            
         }
         
         load=0;

      } else if (listptr->type) {
         
         /* Change to new directory */
         if ((EXCLUDE && match_exclude(get_cur_dir(listptr,dirstr),listptr->name) == 1) || (chdir(listptr->name) == -1)) {
            load=0;
            list_down=1;
         } else {
            
            close(dir_fd);
            
            if ((dir_fd = open(".",O_RDONLY)) == -1) {
               fprintf(stderr,"%s: open: CDB_5: %s\n",progname,strerror(errno));
               return(1);
            }
         }
                  
      }

   }   
   
   rename(TMPSLOCATEDB,SLOCATEDB);
   chmod(SLOCATEDB,00640);
   chown(SLOCATEDB,UID,SLOC_GID);   
   return res;
}

/* Decode Database */

int
decode_db(char *str)
{
   int res = 1;
   FILE *fd;
   short code_num;
	int pathlen=0;
   register char ch;
   int grow=0;
   int jump=0;
	int first=1;
	char *codedpath=NULL;
	char *ptr;
   int printit;
   struct stat statres;	
  
   if ((fd = fopen(SLOCATEDB,"r")) == NULL) {
      fprintf(stderr,"%s: decode_db(): open: (%d) %s\n",progname,errno,strerror(errno));
      return(0);
   }

   slevel = getc(fd);
	
	codedpath = malloc(MIN_BLK);
	codedpath[0] = '\0';
   ptr=codedpath;
   while ((code_num = getc(fd)) != EOF) {
		
      if (code_num == SLOC_ESC) {
         code_num = get_short(fd);
      } else if (code_num > 127)
				 code_num = code_num - 256;
		
      jump=0;
		
		if (code_num < 0)
			grow += code_num;
		
		ptr+=code_num;
		
		pathlen = ptr - codedpath;

      while(!jump) {
         ch = getc(fd);
         grow++;
			pathlen++;
			if (grow == 64) {
				codedpath = realloc(codedpath,pathlen+MIN_BLK);
				grow=0;
			}
         codedpath[pathlen-1] = ch;
	
         if (ch == '\0')
				jump=1;
      }
		
		if (first) {
			ptr=ptr+strlen(codedpath);
			first=0;
		}
		
		pathlen--;

      if (strstr(codedpath,str) != NULL) {
         if (slevel == '1') {
				
            if (UID == 0 || lstat(codedpath,&statres) == 0)
					 printit=1;
				else
					 printit=0;
         } else
             printit=1;
         
         if (printit)
             printf("%s\n",codedpath);
      }
   }      
   
   fclose(fd);
   
   return(res);
}

/* Main Function */

int
main(int args, char **argv)
{
   int res=0;
   int ch;
   extern char *optarg;
   extern int optind, opterr, optopt;
   char *p;
   int SPECDIR=0;
   int ROOTDIR=0;   
   char *spec_dir=NULL;
   int optlen;

   progname = ((p = rindex(argv[0],'/')) ? p+1 : *argv);   
   
   if (args < 2) 
       usage();

	if (!check_dir())
		 return(1);	
	
   UID = getuid();        
   GID = getgid();      
   
   while ((ch = getopt(args,argv,"VuU:e:l:")) != EOF) {
      switch(ch) {
      case 'V':
         printf("%s",VERSION);
         exit(1);
         break;
      case 'e':
         EXCLUDE_DIR = parse_exclude(optarg);
         break;
      case 'l':
         slevel = optarg[0];
         if (slevel != '0' && slevel != '1') {
            fprintf(stderr,"%s: ERROR: Security level must be 0 or 1.\n",progname);
            exit(1);
         }
         break;
      case 'u':
         ROOTDIR=1;         
         break;
      case 'U':
         SPECDIR=1;
         optlen=strlen(optarg);
         spec_dir = malloc(optlen+1);
         memmove(spec_dir,optarg,optlen);
         spec_dir[optlen] = '\0';         
         break;
      default:
         return(1);
         break;
         
      }
   }
   
   SLOC_GID = get_gid(GRPFILE);
   
   if (SPECDIR) {
      res = create_db(spec_dir);
   }
   else if (ROOTDIR)
      res = create_db((char *)NULL);
   else if (argv[optind]) {
      res = decode_db(argv[1]);
   } else
       usage();

   if (!res)
       return(1);
   
   return(0);
}
